from domain.plan.infrastructure.registered_layer_processor.helper import fetch_point, fetch_point_id, \
    check_key_exists_arc, check_key_exists, check_obj_len, type_lookup
from domain.plan.model.point_detail import PointDetail
from domain.plan.model.parcel import Parcel
from domain.plan.model.parcel_line import ParcelLine
from domain.plan.model.admin import Admin as PlanAdmin
from domain.plan.model.line_detail import LineDetail
from shapely.geometry.multipolygon import MultiPolygon
from shapely import wkt
import logging


class LoadGeometryInfoToPlanDB:
    """
    This class helps to load the geometry info to admin, line_detail, point, point_detail, parcel,
    parcel_line tables.
    """

    def __init__(self, plan_db_session, event_id, file_geometry, document_version_id):
        self.plan_db = plan_db_session
        self.event_id = event_id
        self.geometry = file_geometry
        self.document_version_id = document_version_id

    def load_all(self):
        lines = []
        point_details = []
        parcels = []
        if self.geometry.lines:
            lines = self.fetch_line_detail(self.geometry.lines, self.geometry.survey_number)
        if self.geometry.polygons:
            parcels = self.load_parcel(self.geometry.polygons, lines)
        if self.geometry.points:
            point_details = self.load_point_detail(self.geometry.points)
        if self.geometry.points:
            points = self.load_point(self.geometry.points)
        if self.geometry.admin:
            lines_admin = check_obj_len(lines)
            parcels_admin = check_obj_len(parcels)
            point_details_admin = check_obj_len(point_details)
            admin_id = self.load_plan_admin(self.geometry.admin, lines_admin, parcels_admin, point_details_admin,
                                         self.document_version_id)
            return admin_id

    # geometries.admin
    def load_plan_admin(self, admin_info, lines, parcels, point_details, document_version_id):
        admin = PlanAdmin(
            line_detail=lines,
            point_detail=point_details,
            parcel=parcels,
            survey_number=admin_info.plan_number,
            survey_date=admin_info.date_of_survey,
            reg_date=admin_info.registration_date,
            eplan=None,
            lga_code=admin_info.lga_code,
            create_event_id=self.event_id,
            excluded=False,
            excluded_reason=None,
            xml_working_folder=admin_info.plan_number + "/" + document_version_id,
            pseudo=False,
            ignore_event_id=None
        )
        with self.plan_db as open_session:
            open_session.add(admin)
            open_session.commit()
            return admin.id

    # geometries.lines
    def fetch_line_detail(self, lines_info, survey_number):
        lines = []
        for line_key in lines_info.keys():
            line_obj = lines_info[line_key]

            line_detail_type_id = type_lookup(self.plan_db, 'line_type', line_obj.line_type)
            line_detail_radius = check_key_exists_arc(line_obj, 'radius')
            line_detail_arc_length = check_key_exists_arc(line_obj, 'arc_length')
            line_detail_rotation = check_key_exists_arc(line_obj, 'rot')
            line_detail_azimuth_accuracy = check_key_exists(line_obj, 'azimuthAccuracy')
            bearing_type_id = type_lookup(self.plan_db, 'observation_type', line_obj.azimuth_type)
            distance_type_id = type_lookup(self.plan_db, 'observation_type', line_obj.distance_type)

            line_detail = LineDetail(
                name=line_obj.name,
                line_type=line_detail_type_id,
                setup_point_id=line_obj.setup_point.associated_point_oid,
                target_point_id=line_obj.target_point.associated_point_oid,
                bearing_dms=line_obj.hp_bearing,
                bearing_type=bearing_type_id,
                distance=line_obj.distance,
                distance_type=distance_type_id,
                radius=line_detail_radius,
                ref_survey_number=survey_number,
                description=line_obj.desc,
                adjust_type=None,
                distance_std=line_obj.distance_std,
                direction_std=line_detail_azimuth_accuracy,
                original_distance=line_obj.orig_distance_std,
                original_bearing_dms=None,
                az_adopt_factor_dms=line_obj.az_adopt_fact,
                unswung_bearing_dms=None,
                line_geometry=line_obj.geometry.wkt,
                rotation=line_detail_rotation,
                arc_length=line_detail_arc_length,
                create_event_id=self.event_id,
                ignore_event_id=None,
                pseudo=False
            )
            lines.append(line_detail)
        return lines

    # geometries.point_detail
    def load_point_detail(self, point_info):
        point_detail_list = []
        for value in point_info.values():
            if value.point_oid is None:
                point_id_up = fetch_point_id(value.geometry.wkt)
            else:
                point_id_up = value.point_oid
            point_detail_mon_conditon = check_key_exists(value, 'mon_conditon')
            point_type_id = type_lookup(self.plan_db, 'point_type', value.point_type)
            point_state_id = type_lookup(self.plan_db, 'point_state', value.point_state)

            point_detail = PointDetail(point_id=point_id_up,
                                       point_type=point_type_id,
                                       point_state=point_state_id,
                                       mark_type=value.mon_type,
                                       mark_state=value.mon_state,
                                       mark_condition=point_detail_mon_conditon,
                                       ref_survey_number=self.geometry.survey_number,
                                       description=value.mon_desc,
                                       point_geometry=value.geometry.wkt,
                                       create_event_id=self.event_id,
                                       pseudo=False)
            point_detail_list.append(point_detail)
        return point_detail_list

    def load_parcel(self, polygons, lines):
        parcels = []
        for parcel_key in polygons:
            parcel_value = polygons[parcel_key]
            # Get parcel lines list
            parcel_line_list = self.load_parcel_line(parcel_value.line_order)
            parcel_type_id = type_lookup(self.plan_db, 'parcel_type', parcel_value.parcel_type)
            parcel_state_id = type_lookup(self.plan_db, 'parcel_state', parcel_value.parcel_state)
            parcel_class_id = type_lookup(self.plan_db, 'parcel_class', parcel_value.parcel_class)
            parcel_format_id = type_lookup(self.plan_db, 'parcel_format', parcel_value.parcel_format)

            polygon_geom = wkt.loads(parcel_value.geometry.wkt)
            try:
                multipolygon_geom = MultiPolygon([polygon_geom])
                parcel_geom_val = multipolygon_geom.wkt
            except Exception:
                parcel_geom_val = None
            parcel = Parcel(spi=parcel_value.name,
                            parcel_type=parcel_type_id,
                            parcel_state=parcel_state_id,
                            parcel_class=parcel_class_id,
                            parcel_format=parcel_format_id,
                            parcel_line=parcel_line_list,
                            description=parcel_value.desc,
                            name=parcel_value.name,
                            parcel_geometry=parcel_geom_val,
                            create_event_id=self.event_id)
            parcels.append(parcel)
        return parcels

    def load_parcel_line(self, line_order):
        parcel_lines = []
        for parcel_line_key in line_order:
            parcel_line_value_list = line_order[parcel_line_key]
            for i in range(len(parcel_line_value_list)):
                value = parcel_line_value_list[i]
                ring = parcel_line_key[parcel_line_key.find("-")+1:]
                parcel_line = ParcelLine(seq=i,
                                         reverse=value.reversed,
                                         ring=ring,
                                         parcel_index=0,
                                         create_event_id=self.event_id
                                         )
                parcel_lines.append(parcel_line)
        return parcel_lines

    # geometries.point_detail
    def load_point(self, point_info):
        point_list = []
        for value in point_info.values():
            if value.point_oid is None:
                point_id_up = fetch_point_id(value.geometry.wkt)
            else:
                point_id_up = value.point_oid

            point = fetch_point(self.plan_db, self.event_id, value, point_id_up)
            if point is not None:
                point_list.append(point)
        with self.plan_db as open_session:
            if len(point_list) > 0:
                open_session.add_all(point_list)
                open_session.commit()
        return point_list
