import os
import platform as pf
from inspect import getsourcefile
import subprocess
from subprocess import PIPE
from pathlib import Path
import warnings

from utilities.dcmgeometrysdk.dna.dnawriters import DNAWriters


class DNARunner:
    def __init__(self, dna_dir, geoid_fp=None, multi_thread=False, max_iter=10,
                 iter_thresh=.0005, output_dir=None, filename=None, cwd=None, docker=False, mount_dir=None,
                 host_directory=None, container_name='DNA'):
        if cwd is None:
            self.cwd = Path(os.path.split(os.path.abspath(getsourcefile(lambda: 0)))[0])
        else:
            self.cwd = cwd
        self.system_type = pf.system()
        if docker is True:
            self.system_type = 'docker'

        self.dna_dir = dna_dir
        self.max_iterations = max_iter
        self.iter_threshold = iter_thresh
        self.output_dir = self.set_output_dir(output_dir)
        self.mount_dir = Path(str(output_dir).replace(str(host_directory), (mount_dir)))
        self.host_directory = host_directory
        self.docker = docker
        if docker is True:
            self.docker_commands = ['/usr/local/bin/docker', 'exec', '-w',
                                    f'{self.mount_dir}', '-i',  f'{container_name}']
            self.docker_prefix = f'/usr/local/bin/docker exec -w {self.mount_dir} -it {container_name} '
        else:
            self.docker_commands = []
            self.docker_prefix = ''

        if filename.endswith('.xml'):
            filename = filename.replace('.xml', '')
        self.filename = filename
        self.prefix = ''
        self.suffix = ''
        self.ext = ''
        self.convert_stn_heights = None
        self.geoid_fp = self.set_geoid_location(geoid_fp)
        self.dms_msr_format = None
        self.sep = ''
        self.multi_thread = multi_thread
        self.set_run_info()

        self.dna_import = self.set_dna_fps('import')
        self.dna_reftran = self.set_dna_fps('reftran')
        self.dna_adjust = self.set_dna_fps('adjust')
        self.dna_plot = self.set_dna_fps('plot')
        self.dna_segment = self.set_dna_fps('segment')
        self.dna_geoid = self.set_dna_fps('geoid')
        self.execution_lines = []
        self.dms_msr_format = 2
        self.adj_type = '.simult.adj'
        self.stn_coord_types = 'ENzPLHh'
        self.output_type = '--output-adj-msr'
        self.cmd_output = None
        self.results_file = None

        # add more dna settings here.

    def set_dna_fps(self, module):
        fp = Path(self.dna_dir, self.prefix + module + self.suffix)
        if fp.exists():
            return fp
        elif self.docker is True:
            return fp
        else:
            pass
            # warnings.showwarning(f'{module} is not present at {fp}')

    def set_output_dir(self, output_dir=None):
        if output_dir is None:
            # warnings.showwarning(f'No output directory set, setting to {self.cwd}')
            return self.cwd
        else:
            return Path(output_dir)

    def set_stns_msrs(self, items):
        if isinstance(items, list):
            if self.docker is True:
                items = [Path(str(stn).replace(str(self.output_dir), str(self.mount_dir))) for stn in items]
            fn = items[0]
            items = [str(i) for i in items]
            item = ' '.join(items)
        else:
            if self.docker is True:
                items = Path(str(items).replace(str(self.output_dir), str(self.mount_dir)))
            fn = items
            item = items
        return fn, item, items

    def build_execution_cmd(self, msr_locations, stn_locations):

        fn, msr_location, msr_locations = self.set_stns_msrs(msr_locations)
        fn, stn_location, stn_locations = self.set_stns_msrs(stn_locations)

        if self.filename is None:
            self.filename = fn.parts[-1].replace('.stn', '')
            # warnings.showwarning(f'No filename set, setting to {self.filename}')
        if self.docker is False:
            lines = [f'cd "{self.output_dir}"']
        else:
            lines = []
        lines.append(f'{self.docker_prefix}"{str(self.dna_import)}" '
                     f'-n "{self.filename}" {str(stn_location)} {str(msr_location)} -r GDA2020')

        if self.geoid_fp is not None:
            if self.convert_stn_heights is True:
                lines.append(f'{self.docker_prefix}'
                             f'"{str(self.dna_geoid)}" -n "{self.filename}" '
                             f'-g "{str(self.geoid_fp)}" --convert-stn-hts')
            else:
                lines.append(f'{self.docker_prefix}"{str(self.dna_geoid)}" '
                             f'-n "{self.filename}" -g "{str(self.geoid_fp)}"')
        else:
            pass
            # warnings.showwarning('No geoid file is set')

        if self.multi_thread is True:
            mt = '--multi-thread'
            lines.append(f'{self.docker_prefix}"{str(self.dna_segment)}" -n "{self.filename}"')
            self.adj_type = '.phased-mt.adj'
        else:
            mt = ''
            self.adj_type = '.simult.adj'

        lines.extend([
            f'{self.docker_prefix}"{str(self.dna_adjust)}" "{self.filename}" {self.output_type} '
            f'--dms-msr-format {str(self.dms_msr_format)} --stn-coord-types {self.stn_coord_types} --stn-corrections '
            f'--max-iterations {self.max_iterations} --iteration-threshold {self.iter_threshold} {mt}'])
        self.execution_lines = lines
        return lines

    def determine_output(self):
        if self.docker is True:
            output_dir = self.mount_dir
        else:
            output_dir = self.output_dir
        return output_dir

    def write_execution_cmd_to_file(self, filename_suffix=''):

        if len(filename_suffix) > 0:
            filename_suffix = '_' + filename_suffix
        self.cmd_output = Path(self.output_dir, self.filename + filename_suffix + self.ext)
        if len(self.execution_lines) > 0:
            with open(self.cmd_output, 'w') as open_txt:
                for line in self.execution_lines:
                    open_txt.write(line + '\n')

    def execute_cmd(self):
        if self.cmd_output is not None:
            if self.cmd_output.exists():
                subprocess.call([self.cmd_output], cwd=self.output_dir)
                self.set_results()

    def run_dna_via_subprocesses(self, msr_locations, stn_locations):
        output_dir = self.output_dir
        if self.docker is True:
            dl = self.docker_commands
        else:
            dl = self.docker_commands
        sub_list = dl + [str(self.dna_import), '-n', self.filename]
        fn, stn_location, stn_locations = self.set_stns_msrs(stn_locations)
        if isinstance(stn_locations, list):
            for item in stn_locations:
                sub_list.append(str(item))
        else:
            sub_list.append(str(stn_locations))
        fn, msr_location, msr_locations = self.set_stns_msrs(msr_locations)
        if isinstance(msr_locations, list):
            for item in msr_locations:
                sub_list.append(str(item))
        else:
            sub_list.append(str(msr_locations))
        sub_list += ['-r', 'GDA2020']
        if self.docker is False:
            cwd = output_dir
        else:
            cwd = output_dir

        p = subprocess.run(sub_list, cwd=cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

        if self.geoid_fp is not None:
            if self.geoid_fp.exists() or self.docker is True:
                if self.convert_stn_heights is True:

                    p = subprocess.run(dl + [str(self.dna_geoid), '-n', self.filename, '-g',
                                       str(self.geoid_fp), '--convert-stn-hts'], cwd=cwd, stdin=PIPE,
                                       stdout=PIPE, stderr=PIPE)
                else:

                    p = subprocess.run(dl + [str(self.dna_geoid), '-n', self.filename, '-g',
                                       str(self.geoid_fp)], cwd=cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        else:
            pass
            # warnings.showwarning('No geoid file is set')

        adj_list = dl + [
            self.dna_adjust, self.filename, self.output_type,
            '--dms-msr-format', self.dms_msr_format, '--stn-coord-types', self.stn_coord_types, '--stn-corrections',
            '--max-iterations', self.max_iterations, '--iteration-threshold', self.iter_threshold, '--multi-thread']
        adj_list = [str(i) for i in adj_list]

        if self.multi_thread is True:
            p = subprocess.run(dl + [str(self.dna_segment), '-n', self.filename],
                               cwd=cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            p = subprocess.run(adj_list, cwd=cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            self.adj_type = '.phased-mt.adj'
        else:
            # QMessageBox.information(None, "Swing Value",
            #                         str(adj_list[:-1]))
            p = subprocess.run(adj_list[:-1], cwd=cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            self.adj_type = '.simult.adj'

        results = self.set_results()
        return results

    def set_results(self):
        results = Path(str(self.output_dir), self.filename + self.adj_type)
        if results.exists is False:
            results = None
            # warnings.showwarning("No results file, adjustment hasn't worked for some reason")
        self.results_file = results
        return results

    def set_geoid_location(self, geoid_fp):
        if geoid_fp is None:
            geoid_fp = Path(self.dna_dir, 'geoid_file', 'AUSGeoid2020_20170908.gsb')
        if geoid_fp.exists() is False:
            geoid_fp = Path(self.dna_dir, 'geoid_file', 'AUSGeoid2020_20170908.gsb')
            if self.docker is False:
                if geoid_fp.exists() is False:
                    pass
                    # warnings.showwarning('Could not find geooid file')
        return geoid_fp

    def set_run_info(self):
        if self.system_type.lower() == 'windows':
            self.ext = '.bat'
            self.sep = '&'
            self.suffix = '.exe'
            # self.directory = 'windows1064'
        elif self.system_type.lower() == 'darwin':
            self.ext = '.sh'
            self.prefix = 'dna'
            # self.directory = 'darwin'
        else:
            self.ext = '.sh'
            self.prefix = 'dna'
            # self.directory = 'linux'


