from pathlib import Path
from datetime import datetime
import boto3
import os
import json
from copy import deepcopy
from itertools import combinations


from utilities.dcmgeometrysdk.geometryfunctions.otherfunctions import sort_by_tuple
from utilities.dcmgeometrysdk.geometryfunctions.conversionsfunctions import dd2hp
from utilities.dcmgeometrysdk.geometryfunctions.transformationfunctions import transform_geoms
from utilities.dcmgeometrysdk.geometryfunctions.bearingdistancefunctions import calc_inside_360
from utilities.dcmgeometrysdk.dna.standard_deviations import StandardDeviations

def get_dms(b1):
    d, ms = str(b1).split('.')
    ms = ms.ljust(4, '0')
    m = ms[:2]
    s = ms[2:4]
    if len(ms) > 4:
        millis = ms[4:6]
    else:
        millis = '0'
    millis = millis.ljust(2, '0')

    return d, m, s, millis


class AprioriValues:
    def __init__(self, name, b, d, ppm):
        self.name = name
        self.bearing_std = b
        self.distance_std = d
        self.ppm = ppm


class DNADirectionSet:
    def __init__(self, ip, to_stations, lines, correlate=False, points=None, angle_std=1):
        self.ip = ip
        self.stations = to_stations
        self.correlate = correlate
        self.lines = lines
        if points is not None:
            self.points = points
        else:
            self.points = {}
        self.line_sets = {}
        self.order_stations()
        self.direction_sets = self.generate_direction_set()
        self.angle_std = angle_std

    def order_stations(self):
        stations = []
        for item in self.stations:
            line1 = self.lines.get((self.ip, item))
            if line1 is not None:
                self.line_sets[(self.ip, item)] = line1
            else:
                line1 = deepcopy(self.lines.get((item, self.ip)))
                line1.flip_direction()
                self.line_sets[(self.ip, item)] = line1
            stations.append((line1, line1.dd_bearing))
        self.stations = [i[0] for i in sorted(stations, key=lambda x: x[1])]

    def generate_direction_set(self):
        directions = []
        c2 = None
        if self.correlate is False:
            dset_length = 1
        else:
            dset_length = len(self.stations) - 1
        for item in self.stations:
            if item.setup_point.name == self.ip:
                target = item.target_point.name
            else:
                target = item.setup_point.name
            order = self.stations.index(item)
            if order == 0:
                c2 = self.points.get(target).ccc
            direction = DNADirection(self.ip, target, dset_length, item, order=order)
            if c2 is False or self.points.get(self.ip).ccc is False or self.points.get(target).ccc is False:
                if self.correlate is False:
                    if len(directions) > 0 and order != 1:
                        directions.append(directions[0])
                    directions.append(direction)
                else:
                    directions.append(direction)
        if len(directions) < 2:
            directions = []
        return directions

    def generate_angles_from_stations(self, important_angles=(0, 90, 180, 270)):
        angles = []
        for a, b in combinations(self.line_sets, 2):
            if a[1] != self.ip and b[1] != self.ip:
                l1 = self.line_sets[a]
                l2 = self.line_sets[b]
                if l1.dd_bearing > l2.dd_bearing:
                    angle = calc_inside_360(l2.dd_bearing - l1.dd_bearing)
                else:
                    angle = l2.dd_bearing - l1.dd_bearing
                #print(round(angle, 2))
                if round(angle, 2) in important_angles:
                    if self.points.get(a[1]).ccc is False or self.points.get(self.ip).ccc is False or\
                            self.points.get(b[1]).ccc is False:
                        angled = DNAAngle(self.ip, a[1], b[1], round(angle, 2), std=self.angle_std)
                        angles.append(angled)
        return angles


class DNAAngle:
    def __init__(self, ip, stn1, stn2, angle, std):
        self.ip = ip
        self.stn2 = stn1
        self.stn3 = stn2
        self.angle = dd2hp(angle, rnd=True)
        self.std = std

    def write_line(self):
        d, m, s, millis = get_dms(self.angle)
        row = ('A'.rjust(1) +
               ''.rjust(1) +
               str(self.ip)[:20].ljust(20) +
               str(self.stn2)[:20].ljust(20) +
               str(self.stn3)[:20].ljust(20) +
               ''.rjust(14) +
               d.rjust(3) + ' ' +
               m.ljust(2) +
               (' ' + s + '.' + millis).ljust(8) +
               str(self.std).rjust(9))
        return row


class DNADirection:
    def __init__(self, ip, to_station, dset_length, line, order):
        self.ip = ip

        self.to_station = to_station
        self.line = line
        std = line.bearing_std
        if not isinstance(std, (float, int)):
            std = 60
        self.std = '{:.4f}'.format(float(std))
        if len(self.std) > 9:
            self.std = self.std[:9]
        self.order = order
        self.dset_length = dset_length

    def write_line(self):
        d, m, s, millis = get_dms(self.line.hp_bearing)
        if self.order == 0:
            c1 = self.ip
            c2 = self.to_station
            c3 = self.dset_length
        else:
            c1 = ''
            c2 = ''
            c3 = self.to_station

        row = ('D'.rjust(1) +
               ''.rjust(1) +
               str(c1)[:20].ljust(20) +
               str(c2)[:20].ljust(20) +
               str(c3)[:20].ljust(20) +
               ''.rjust(14) +
               d.rjust(3) + ' ' +
               m.ljust(2) +
               (' ' + s + '.' + millis).ljust(8) +
               self.std.rjust(9))

        return row

class DNADistances:
    def __init__(self, line, ins_height=0, targ_height=0):
        self.distance = line.distance
        self.dist_std = line.distance_std
        # if self.dist_std > self.distance and self.distance < 100:
        #     self.dist_std = self.distance * 2
        self.setup = line.setup_point.name
        self.target = line.target_point.name
        self.ins_height = ins_height
        self.targ_height = targ_height

    def write_line(self):
        line = ('S'.rjust(1)
                + ''.ljust(1)
                + str(self.setup)[:20].ljust(20)
                + str(self.target)[:20].ljust(20)
                + ''.ljust(20)
                + ('{:.4f}'.format(float(self.distance))).ljust(14)
                + ''.ljust(14)
                + ('{:.4f}'.format(float(self.dist_std)).rjust(9))
                + ('{:.4f}'.format(float(self.ins_height)).rjust(7))
                + ('{:.4f}'.format(float(self.targ_height)).rjust(7)))
        return line


class DNABearing:
    def __init__(self, line, std):
        self.setup = line.setup_point.name
        self.target = line.target_point.name
        self.bearing = line.hp_bearing
        self.std = std

    def write_line(self):
        d, m, s, millis = get_dms(self.bearing)
        line = ('B'.rjust(1)
                + ''.ljust(1)
                + str(self.setup)[:20].ljust(20)
                + str(self.target)[:20].ljust(20)
                + str('')[:20].ljust(20) +
                ''.rjust(14) +
                d.rjust(3) + ' ' +
                m.ljust(2) +
                (' ' + s + '.' + millis).ljust(8) +
                str(self.std).rjust(9))
        return line



class DNAWriters:
    def __init__(self, geometries,  output_dir=Path(), filename=None, out_datum=7844, survey_year=None,
                 survey_type='surveyed', constrained_marks=None, ignored_stns=None, profile_location=None,
                 profile='dcmvr', stds=None):
        geometries = deepcopy(geometries)
        self.direction_sets = []
        self.distances = []
        self.bearing_constraints = []
        self.points = geometries.points  # dict of PointGeom items taken from the dcmgeometry class
        self.lines = geometries.lines  # dict of bearings and distances from LineGeom dcmgeometry class
        if filename is None:
            filename = geometries.survey_number
        self.filename = str(filename)
        self.out_path = output_dir
        self.in_datum = geometries.crs
        if self.in_datum is None:
            self.in_datum = 7855
        self.out_datum = out_datum
        self.exclude_lines = set()
        if survey_year is not None:
            self.survey_year = survey_year
        else:
            self.survey_year = geometries.survey_year
        self.survey_type = survey_type
        self.ignored_obs = self.get_ignored_obs(ignored_stns)
        self.constrained_marks = None
        self.set_constrained(constrained_marks)
        self.min_distance = None
        self.bearing_std = None
        self.include_bearings = self.set_bearing_constraints()
        self.correlate = self.set_correlate()
        self.angle_std = None
        self.include_angles = self.set_include_angles()
        if stds is None:
            stds = StandardDeviations(profile, profile_location, self.survey_year)
        self.lines = stds.set_line_stds(self.lines)

    def set_correlate(self, correlate=False):
        """this sets the dynadjust grouping of directions to correlate so that when the file is written you will only
        get one set of directions per set rather than all directions available for that station in a group"""
        self.correlate = correlate
        return correlate

    def set_include_angles(self, include_angles=False, angle_std=1):
        self.include_angles = include_angles
        self.angle_std = angle_std
        return include_angles

    def set_bearing_constraints(self, include_bearings=False, distance=.2, std=324000):
        self.include_bearings = include_bearings
        self.min_distance = distance
        self.bearing_std = std
        return include_bearings

    def get_ignored_obs(self, ignored_obs=None):
        # TODO this needs testing
        """this is used if there are ignored stations, these will not be used in
        the adjustment """
        new_ignored = set()

        if ignored_obs is not None:
            new_ignored = {i for i in ignored_obs}
        else:
            ignored_obs = set()
        used = set()
        for k, v in self.lines.items():
            if v.azimuth_type in ['Ignored'] or v.distance_type in ['Ignored']:
                if k[0] not in used:
                    new_ignored.add(k[0])
                if k[1] not in used:
                    new_ignored.add(k[1])
            else:
                used.add(k[0])
                used.add(k[1])
                if k[0] in new_ignored and k[0] not in ignored_obs:
                    new_ignored.remove(k[0])
                if k[1] in new_ignored and k[1] not in ignored_obs:
                    new_ignored.remove(k[1])

        return new_ignored

    def set_constrained(self, constrained_marks=None):
        """sets the constrained mark names that the user want to make constrained (CCC)"""
        if constrained_marks is None:
            constrained_marks = set()
            for k, v in self.points.items():
                if v.ccc is True:
                    constrained_marks.add(k)
            self.constrained_marks = constrained_marks
        elif isinstance(constrained_marks, set):
            self.constrained_marks = constrained_marks
        return constrained_marks

    def create_bearing_constraints(self, refresh_bearings=True):
        if refresh_bearings is True:
            self.bearing_constraints = []
        lines = []
        for line in self.lines.values():
            if line.distance < self.min_distance:# and line.setup_point.ccc is False and line.target_point.ccc is False:
                b = DNABearing(line, self.bearing_std)
                self.bearing_constraints.append(b)
                lines.append(b.write_line())
        return lines

    def set_exclude(self, line_names=None):
        """manually exclude some names from the adjustment, these are set here with a set of line names"""
        if line_names is None:
            line_names = set()
        self.exclude_lines = line_names

    def create_stn_header(self):
        """
        creates the dna station header line, this contains the header information required by dynadjust to
        run a minimally constrained adjustment
        :param number_of_stations: the number of stations that are going to be written into the file
        :return: stn: the station header
        """
        # dt1 = '01.01.1994'
        if self.out_datum == 7844:
            g = 'GDA'
        else:
            g = 'GDA'

        dt = datetime.strftime(datetime.now(), '%d.%m.%Y')
        header = ('!#=DNA'.rjust(6) +
                  '3.01'.rjust(6) +
                  'STN'.rjust(3) +
                  dt.rjust(14) +
                  f'{g}2020'.rjust(14) +
                  '01.01.2020'.rjust(14) +
                  (str(len(self.points))).rjust(10))
        return [header]

    def create_stn_lines(self):
        """
        creates the dna station file data lines, is a list that contains pre-formatted strings that will be written to
        the stn file, setting the elevation at zero for all items
        :return: a list of items to be writen as lines in the dynadjust stn file.
        """
        stn = []
        elevation = '0'
        projection = 'UTM'
        zone = str(self.in_datum)[-2:]
        if self.in_datum is None:
            self.in_datum = 7855
        if self.out_datum == 7844:
            projection = 'LLH'
            zone = ''

        if self.in_datum != self.out_datum:
            self.points = transform_geoms(self.points, self.in_datum, self.out_datum)
            self.in_datum = self.out_datum

        if self.constrained_marks is None:
            pnts = {pt: v for pt, v in self.points.items() if v.point_type == 'sideshot' and pt not in self.ignored_obs}
            constrained = set(sorted(pnts.keys())[0])
        else:
            constrained = self.constrained_marks

        pnts = {pt: v for pt, v in self.points.items() if pt not in self.ignored_obs}
        for pt, val in pnts.items():

            if pt in constrained:
                const = 'CCC'
            else:
                const = 'FFC'

            x = val.geometry.x
            y = val.geometry.y
            # set to vic
            if x < 100000 and projection != 'LLH':
                x += 250000
                y += 5900000

            if projection == 'LLH':
                y = dd2hp(float(y))
                x = dd2hp(float(x))
                e = "{0:.9f}".format(y).rjust(20)
                n = "{0:.9f}".format(x).rjust(20)
            else:
                e = "{0:.3f}".format(float(x)).rjust(20)
                n = "{0:.3f}".format(float(y)).rjust(20)

            line = (str(pt)[:20].ljust(20)  # Pt ID
                    + const.ljust(3)
                    + ''.ljust(1)  # Constraint
                    + projection.ljust(3)  # Projection
                    + e  # Easting
                    + n  # Northing
                    + elevation.rjust(20)  # Elevation
                    + zone.rjust(3))  # Hemisphere/Zone input
            if const == 'CCC':
                stn = [line] + stn
            else:
                stn.append(line)
        return stn

    def write_stn_file_encode(self):
        """
        creates a binary linestring for dna stn format groups the measurements and adds the estimates that can be written
        to a file in AWS
        :return:
        """
        lines = self.create_stn_header()
        lines += self.create_stn_lines()

        line_string = ''
        for line in lines:
            line_string += line + "\n"

        line_string = line_string.encode('UTF-8')
        return line_string

    def write_stn_file(self, filename=None, out_path=None):
        """Writes the station file for Dynadjust"""
        if out_path is None:
            out_path = Path(self.out_path)
        else:
            out_path = Path(out_path)
        if filename is None:
            filename = self.filename

        out_path.mkdir(parents=True, exist_ok=True)
        out_path = Path(out_path, str(filename) + '.stn')
        lines = self.write_stn_file_encode()
        lines = lines.decode('UTF-8')
        with open(out_path, 'w') as stn_file:
            stn_file.write(lines)
        return out_path

    def group_obs(self):
        """groups all the observations in the file to the setup points"""
        obs = {}
        dist_obs = {}

        all_obs = {}
        for k, v in self.lines.items():
            all_obs[k[0], k[1]] = v
            x = deepcopy(v)
            x.flip_direction()
            all_obs[k[1], k[0]] = x
        for k, v in all_obs.items():
            if v.line_type not in self.exclude_lines:
                ignore_types = ['Ignored', 'closure']
                if v.azimuth_type not in ignore_types:
                    exist_obs = obs.get(k[0], [])
                    exist_obs.append((k[1]))
                    obs[k[0]] = exist_obs
                if v.distance != 0:
                    if v.distance_type not in ignore_types and v.reversed is False:
                        dist_obs[k] = v
        obs = {k: v for k, v in obs.items() if len(v) > 1}
        return obs, dist_obs

    def create_msr_dir_lines(self, obs, refresh_dir_set=True):
        if refresh_dir_set is True:
            self.direction_sets = []
        """creates the list of measurement direction lines to write into the measurement Dynadjust file"""
        # line_lookups = {k:v for k,v in line_lookups.items() if v.get('bearing_type' != 'Ignored')}
        lines = []

        for k, v in obs.items():
            # sort to get clockwise angle order

            d = DNADirectionSet(k, to_stations=v, lines=self.lines, correlate=self.correlate, points=self.points,
                                angle_std=self.angle_std)
            self.direction_sets.append(d)
            for direction in d.generate_direction_set():
                lines.append(direction.write_line())

        return lines

    def create_msr_dist_lines(self, dist_obs, ins_height=0, targ_height=0, refresh_distances=True):
        """creates a list of measurement distance lines for the dynadjust file"""
        if refresh_distances is True:
            self.distances = []
        lines = []
        for k, v in dist_obs.items():
            line = DNADistances(v, ins_height, targ_height)
            self.distances.append(line)
            lines.append(line.write_line())
        return lines

    def create_msr_p_q_constraints(self):
        def create_line_text(c, station, value, std):
            line = (c.rjust(1) + ''.ljust(1) + str(station)[:20].ljust(20) + ''.ljust(40) +
                    str(value)[:14].ljust(14) + ''.ljust(14) + str(std)[:9].ljust(9))
            return line
        lines = []
        for k, v in self.points.items():
            if v.p_constrained >= 0:
                line = create_line_text('P', v.name, v.latitude, v.p_constrained)
                lines.append(line)
            if v.q_constrained >= 0:
                line = create_line_text('Q', v.name, v.longitude, v.q_constrained)
                lines.append(line)
        return lines

    @staticmethod
    def create_msr_header(obs, dist_obs):
        """Creates the measurement file header needed for a dynadjust file"""
        obscount = 0
        for item in obs.values():
            obscount += len(item) - 1
        obscount += len(dist_obs)
        now = datetime.now()
        date = now.strftime('%d.%m.%Y')
        date2 = '01.01.2020'
        g = 'GDA'
        header = ('!#=DNA'.rjust(6) +
                  '3.01'.rjust(6) +
                  'MSR'.rjust(3) +
                  date.rjust(14) +
                  f'{g}2020'.rjust(14) +
                  date2.rjust(14) +
                  str(obscount).rjust(10))

        return [header]

    def write_msr_file_encode(self):
        """
        creates a binary linestring for dna msr format groups the measurements and adds the estimates
        that can be written to a file in AWS
        """

        obs, dist_obs = self.group_obs()
        lines = self.create_msr_header(obs, dist_obs)
        lines += self.create_msr_dir_lines(obs)
        lines += self.create_msr_dist_lines(dist_obs)
        if self.include_angles is True:
            for f in self.direction_sets:
                angles = f.generate_angles_from_stations()
                for a in angles:
                    lines.append(a.write_line())
        if self.include_bearings is True:
            lines += self.create_bearing_constraints()

        lines += self.create_msr_p_q_constraints()
        line_string = ''
        for line in lines:
            line_string += line + "\n"

        line_string = line_string.encode('UTF-8')
        return line_string

    def write_msr_file(self, filename=None, out_path=None):
        if out_path is None:
            out_path = Path(self.out_path)
        else:
            out_path = Path(out_path)
        if filename is None:
            filename = self.filename

        out_path.mkdir(parents=True, exist_ok=True)
        out_path = Path(out_path, str(filename) + '.msr')
        lines = self.write_msr_file_encode()
        lines = lines.decode('UTF-8')
        with open(out_path, 'w') as msr_file:
            msr_file.write(lines)
        return out_path

    def write_stn_msr_encode(self):
        """writes out both station and measurement files to a variable rather than file"""
        stn_lines = self.write_stn_file_encode()
        msr_lines = self.write_msr_file_encode()
        return stn_lines, msr_lines

    def write_stn_msr_file(self, filename=None, out_path=None):
        """writes out both station and measurement files"""
        stn_file = self.write_stn_file(filename, out_path)
        msr_file = self.write_msr_file(filename, out_path)
        return stn_file, msr_file

    def write_stn_msr_to_bucket(self, bucket_name=None, key_loc=''):
        if bucket_name is not None:
            stn_lines, msr_lines = self.write_stn_msr_encode()
            s3r = boto3.resource('s3')

            slog = os.path.join(key_loc, self.filename + '.stn')
            obj = s3r.Object(bucket_name, slog)
            obj.put(Body=stn_lines)

            mlog = os.path.join(key_loc, self.filename + '.msr')
            obj = s3r.Object(bucket_name, mlog)
            obj.put(Body=msr_lines)

            return self.filename + '.stn', self.filename + '.msr'

    def write_stn_geopackage(self, geopackage_name=None):
        pass

    def write_msr_geopackage(self, geopackage_name=None):
        pass